@@ -62,7 +62,7 @@ my_make_filename (const char *first_part, const char *second_part)
       && (home = getenv("HOME")) && *home )
     n += strlen(home);
 
-  name = gcry_xmalloc(n);
+  name = _gcry_xmalloc(n);
   p = (home
        ? stpcpy (stpcpy (name, home), first_part+1 )
        : stpcpy (name, first_part) );
@@ -129,10 +129,10 @@ _gcry_rndegd_set_socket_name (const char *name)
   newname = my_make_filename (name, NULL);
   if (strlen (newname)+1 >= sizeof addr.sun_path)
     {
-      gcry_free (newname);
+      xfree (newname);
       return gpg_error_from_syserror ();
     }
-  gcry_free (user_socket_name);
+  xfree (user_socket_name);
   user_socket_name = newname;
   return 0;
 }
@@ -161,7 +161,7 @@ _gcry_rndegd_connect_socket (int nofail)
 #endif
   if (user_socket_name)
     {
-      name = gcry_strdup (user_socket_name);
+      name = _gcry_strdup (user_socket_name);
       if (!name)
         {
           if (!nofail)
@@ -195,7 +195,7 @@ _gcry_rndegd_connect_socket (int nofail)
       close (fd);
       fd = -1;
     }
-  gcry_free(name);
+  xfree (name);
   if (fd != -1)
     egd_socket = fd;
   return fd;
