@@ -35,7 +35,7 @@
      WORDS_BIGENDIAN       Defined to 1 on big endian systems.
      inline                If defined, it should yield the keyword used
                            to inline a function.
-     HAVE_U32_TYPEDEF      Defined if the u32 type is available.
+     HAVE_TYPE_U32         Defined if the u32 type is available.
      SIZEOF_UNSIGNED_INT   Defined to the size in bytes of an unsigned int.
      SIZEOF_UNSIGNED_LONG  Defined to the size in bytes of an unsigned long.
 
@@ -47,7 +47,7 @@
 
 #ifdef STANDALONE
 #include <stdint.h>
-#define HAVE_U32_TYPEDEF 1
+#define HAVE_TYPE_U32 1
 typedef uint32_t u32;
 #define VERSION "standalone"
 /* For GCC, we can detect endianness.  If not GCC, please define manually.  */
@@ -78,7 +78,7 @@ typedef uint32_t u32;
 
 
 
-#ifndef HAVE_U32_TYPEDEF
+#ifndef HAVE_TYPE_U32
 # undef u32 /* Undef a possible macro with that name.  */
 # if SIZEOF_UNSIGNED_INT == 4
    typedef unsigned int u32;
@@ -87,7 +87,7 @@ typedef uint32_t u32;
 # else
 #  error no typedef for u32
 # endif
-# define HAVE_U32_TYPEDEF
+# define HAVE_TYPE_U32
 #endif
 
 
