@@ -35,7 +35,8 @@
 /* The name of a file used to globally disable selected features. */
 #define HWF_DENY_FILE "/etc/gcrypt/hwf.deny"
 
-/* A table to map hardware features to a string.  */
+/* A table to map hardware features to a string.
+ * Note: Remember to add new HW features to 'doc/gcrypt.texi'.  */
 static struct
 {
   unsigned int flag;
@@ -60,6 +61,7 @@ static struct
     { HWF_INTEL_FAST_VPGATHER, "intel-fast-vpgather" },
     { HWF_INTEL_RDTSC,         "intel-rdtsc" },
     { HWF_INTEL_SHAEXT,        "intel-shaext" },
+    { HWF_INTEL_VAES_VPCLMUL,  "intel-vaes-vpclmul" },
 #elif defined(HAVE_CPU_ARCH_ARM)
     { HWF_ARM_NEON,            "arm-neon" },
     { HWF_ARM_AES,             "arm-aes" },
